# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from argparse import ArgumentParser, RawDescriptionHelpFormatter

from magnopy import __version__
from magnopy._package_info import _warranty, logo


def main():
    parser = ArgumentParser(
        description=logo() + f"\n\nAvailable scripts are:\n\n"
        "* magnopy-lswt\n\n"
        "To call for help for each script type <script name> --help\n"
        "Information below is relevant only to 'magnopy' command.",
        formatter_class=RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "commands",
        default=None,
        help="command/commands on what to do. Use to display some information about package.",
        choices=["logo", "warranty"],
        nargs="*",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="store_true",
        help="print version of magnopy",
    )

    args = parser.parse_args()

    if args.version:
        print(f"magnopy v{__version__}")

    if len(args.commands) == 0:
        parser.print_help()
        return

    for command in args.commands:
        if command == "logo":
            print(logo())
        elif command == "warranty":
            print("\n" + _warranty() + "\n")
        else:
            raise ValueError(f"Sub-command {args.command} is not recognized.")


if __name__ == "__main__":
    main()



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


R"""
Convention of spin Hamiltonian
"""

from magnopy._exceptions import ConventionError
from magnopy.constants._spinham_conventions import _CONVENTIONS

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


class Convention:
    R"""
    Convention of the spin Hamiltonian.

    For the detailed description of the convention problem see :ref:`user-guide_theory-behind_convention-problem`.

    Parameters
    ----------
    multiple_counting : bool, optional
        Whether the pairs of spins are counted multiple times in the Hamiltonian's sums.
    spin_normalized : bool, optional
        Whether spin vectors/operators are normalized to 1. If ``True``, then spin
        vectors/operators are normalized.
    c1 : float, optional
        Numerical factor before the (one spin & one site) term of the Hamiltonian.
    c21 : float, optional
        Numerical factor before the (two spins & one site) term of the Hamiltonian.
    c22 : float, optional
        Numerical factor before the (two spins & two sites) term of the Hamiltonian.
    c31 : float, optional
        Numerical factor before the (three spins & one site) term of the Hamiltonian.
    c32 : float, optional
        Numerical factor before the (three spins & two sites) term of the Hamiltonian.
    c33 : float, optional
        Numerical factor before the (three spins & three sites) term of the Hamiltonian.
    c41 : float, optional
        Numerical factor before the (four spins & one site) term of the Hamiltonian.
    c421 : float, optional
        Numerical factor before the (four spins & two sites & 1+3) term of the Hamiltonian.
    c422 : float, optional
        Numerical factor before the (four spins & two sites & 2+2) term of the Hamiltonian.
    c43 : float, optional
        Numerical factor before the (four spins & three sites) term of the Hamiltonian.
    c44 : float, optional
        Numerical factor before the (four spins & four sites) term of the Hamiltonian.
    name : str, default "custom"
        A label for the convention. Any string, case-insensitive.

    Examples
    --------

    .. doctest::

        >>> from magnopy import Convention
        >>> n1 = Convention(True, True, c21=1, c22=-0.5)
        >>> n2 = Convention(False, True, c21=1, c22=-0.5)
        >>> n3 = Convention(False, True, c22=-0.5)
        >>> n1.multiple_counting
        True
        >>> n1 == n2
        False
        >>> n3.c21
        Traceback (most recent call last):
        ...
        magnopy._exceptions.ConventionError: Convention of spin Hamiltonian has an undefined property 'c21':
        custom convention where
          * Bonds are counted once in the sum;
          * Spin vectors are normalized to 1;
          * Undefined c1 factor;
          * Undefined c21 factor;
          * c22 = -0.5;
          * Undefined c31 factor;
          * Undefined c32 factor;
          * Undefined c33 factor;
          * Undefined c41 factor;
          * Undefined c421 factor;
          * Undefined c422 factor;
          * Undefined c43 factor;
          * Undefined c44 factor.
        >>> n3.name
        'custom'

    """

    __slots__ = (
        "_multiple_counting",
        "_spin_normalized",
        "_c1",
        "_c21",
        "_c22",
        "_c31",
        "_c32",
        "_c33",
        "_c41",
        "_c421",
        "_c422",
        "_c43",
        "_c44",
        "_name",
    )

    def __init__(
        self,
        multiple_counting: bool = None,
        spin_normalized: bool = None,
        c1: float = None,
        c21: float = None,
        c22: float = None,
        c31: float = None,
        c32: float = None,
        c33: float = None,
        c41: float = None,
        c421: float = None,
        c422: float = None,
        c43: float = None,
        c44: float = None,
        name: str = "custom",
    ) -> None:
        if multiple_counting is not None:
            self._multiple_counting = bool(multiple_counting)
        else:
            self._multiple_counting = None

        if spin_normalized is not None:
            self._spin_normalized = bool(spin_normalized)
        else:
            self._spin_normalized = None

        if c1 is not None:
            self._c1 = float(c1)
        else:
            self._c1 = None

        if c21 is not None:
            self._c21 = float(c21)
        else:
            self._c21 = None

        if c22 is not None:
            self._c22 = float(c22)
        else:
            self._c22 = None

        if c31 is not None:
            self._c31 = float(c31)
        else:
            self._c31 = None

        if c32 is not None:
            self._c32 = float(c32)
        else:
            self._c32 = None

        if c33 is not None:
            self._c33 = float(c33)
        else:
            self._c33 = None

        if c41 is not None:
            self._c41 = float(c41)
        else:
            self._c41 = None

        if c421 is not None:
            self._c421 = float(c421)
        else:
            self._c421 = None

        if c422 is not None:
            self._c422 = float(c422)
        else:
            self._c422 = None

        if c43 is not None:
            self._c43 = float(c43)
        else:
            self._c43 = None

        if c44 is not None:
            self._c44 = float(c44)
        else:
            self._c44 = None

        self._name = str(name).lower()

    def summary(self, return_as_string=False):
        r"""
        Gives human-readable summary of the convention.

        Parameters
        ----------
        return_as_string : bool, default False
            Whether to print or return a ``str``. If ``True``, then return an ``str``.
            If ``False``, then print it.

        Examples
        --------

        .. doctest::

            >>> from magnopy import Convention
            >>> n1 = Convention(True, True, c21=1, c22=-0.5)
            >>> n1.summary()
            custom convention where
              * Bonds are counted multiple times in the sum;
              * Spin vectors are normalized to 1;
              * Undefined c1 factor;
              * c21 = 1.0;
              * c22 = -0.5;
              * Undefined c31 factor;
              * Undefined c32 factor;
              * Undefined c33 factor;
              * Undefined c41 factor;
              * Undefined c421 factor;
              * Undefined c422 factor;
              * Undefined c43 factor;
              * Undefined c44 factor.
        """

        summary = [f"{self.name} convention where"]

        if self._multiple_counting is None:
            summary.append("  * Undefined multiple counting;")
        elif self._multiple_counting:
            summary.append("  * Bonds are counted multiple times in the sum;")
        else:
            summary.append("  * Bonds are counted once in the sum;")

        if self._spin_normalized is None:
            summary.append("  * Undefined spin normalization;")
        elif self._spin_normalized:
            summary.append("  * Spin vectors are normalized to 1;")
        else:
            summary.append("  * Spin vectors are not normalized;")

        # One spin
        if self._c1 is None:
            summary.append("  * Undefined c1 factor;")
        else:
            summary.append(f"  * c1 = {self._c1};")

        # Two spins
        if self._c21 is None:
            summary.append("  * Undefined c21 factor;")
        else:
            summary.append(f"  * c21 = {self._c21};")

        if self._c22 is None:
            summary.append("  * Undefined c22 factor;")
        else:
            summary.append(f"  * c22 = {self._c22};")

        # Three spins
        if self._c31 is None:
            summary.append("  * Undefined c31 factor;")
        else:
            summary.append(f"  * c31 = {self._c31};")

        if self._c32 is None:
            summary.append("  * Undefined c32 factor;")
        else:
            summary.append(f"  * c32 = {self._c32};")

        if self._c33 is None:
            summary.append("  * Undefined c33 factor;")
        else:
            summary.append(f"  * c33 = {self._c33};")

        # Four spins
        if self._c41 is None:
            summary.append("  * Undefined c41 factor;")
        else:
            summary.append(f"  * c41 = {self._c41};")

        if self._c421 is None:
            summary.append("  * Undefined c421 factor;")
        else:
            summary.append(f"  * c421 = {self._c421};")

        if self._c422 is None:
            summary.append("  * Undefined c422 factor;")
        else:
            summary.append(f"  * c422 = {self._c422};")

        if self._c43 is None:
            summary.append("  * Undefined c43 factor;")
        else:
            summary.append(f"  * c43 = {self._c43};")

        if self._c44 is None:
            summary.append("  * Undefined c44 factor.")
        else:
            summary.append(f"  * c44 = {self._c44}.")

        summary = ("\n").join(summary)

        if return_as_string:
            return summary

        print(summary)

    @property
    def name(self) -> str:
        r"""
        A label for the convention. Any string, case-insensitive.
        """

        return self._name

    @name.setter
    def name(self, new_value: str):
        self._name = str(new_value).lower()

    ################################################################################
    #                               Multiple counting                              #
    ################################################################################
    @property
    def multiple_counting(self) -> bool:
        r"""
        Whether the pairs of spins are counted multiple times in the Hamiltonian's sums.

        If ``True``, then pairs are counted multiple times.
        """
        if self._multiple_counting is None:
            raise ConventionError(convention=self, property="multiple_counting")
        return self._multiple_counting

    @multiple_counting.setter
    def multiple_counting(self, new_value: bool):
        raise AttributeError(
            "It is intentionally forbidden to set properties of convention. "
            "Use correct methods of SpinHamiltonian class to change convention."
        )

    ################################################################################
    #                            Normalization of spins                            #
    ################################################################################
    @property
    def spin_normalized(self) -> bool:
        r"""
        Whether spin vectors/operators are normalized to 1.

        If ``True``, then spin vectors/operators are normalized.
        """
        if self._spin_normalized is None:
            raise ConventionError(convention=self, property="spin_normalized")
        return self._spin_normalized

    @spin_normalized.setter
    def spin_normalized(self, new_value: bool):
        raise AttributeError(
            "It is intentionally forbidden to set properties of convention. "
            "Use correct methods of SpinHamiltonian class to change convention."
        )

    ################################################################################
    #                                   One spin                                   #
    ################################################################################
    @property
    def c1(self) -> float:
        r"""
        Numerical factor before the (one spin & one site) sum of the Hamiltonian.
        """
        if self._c1 is None:
            raise ConventionError(convention=self, property="c1")
        return self._c1

    @c1.setter
    def c1(self, new_value: float):
        raise AttributeError(
            "It is intentionally forbidden to set properties of convention. "
            "Use correct methods of SpinHamiltonian class to change convention."
        )

    ################################################################################
    #                                   Two spins                                  #
    ################################################################################
    @property
    def c21(self) -> float:
        r"""
        Numerical factor before the (two spins & one site) sum of the Hamiltonian.
        """
        if self._c21 is None:
            raise ConventionError(convention=self, property="c21")
        return self._c21

    @c21.setter
    def c21(self, new_value: float):
        raise AttributeError(
            "It is intentionally forbidden to set properties of convention. "
            "Use correct methods of SpinHamiltonian class to change convention."
        )

    @property
    def c22(self) -> float:
        r"""
        Numerical factor before the (two spins & two sites) sum of the Hamiltonian.
        """
        if self._c22 is None:
            raise ConventionError(convention=self, property="c22")
        return self._c22

    @c22.setter
    def c22(self, new_value: float):
        raise AttributeError(
            "It is intentionally forbidden to set properties of convention. "
            "Use correct methods of SpinHamiltonian class to change convention."
        )

    ################################################################################
    #                                  Three spins                                 #
    ################################################################################
    @property
    def c31(self) -> float:
        r"""
        Numerical factor before the (three spins & one site) sum of the Hamiltonian.
        """
        if self._c31 is None:
            raise ConventionError(convention=self, property="c31")
        return self._c31

    @c31.setter
    def c31(self, new_value: float):
        raise AttributeError(
            "It is intentionally forbidden to set properties of convention. "
            "Use correct methods of SpinHamiltonian class to change convention."
        )

    @property
    def c32(self) -> float:
        r"""
        Numerical factor before the (three spins & two sites) sum of the Hamiltonian.
        """
        if self._c32 is None:
            raise ConventionError(convention=self, property="c32")
        return self._c32

    @c32.setter
    def c32(self, new_value: float):
        raise AttributeError(
            "It is intentionally forbidden to set properties of convention. "
            "Use correct methods of SpinHamiltonian class to change convention."
        )

    @property
    def c33(self) -> float:
        r"""
        Numerical factor before the (three spins & three sites) sum of the Hamiltonian.
        """
        if self._c33 is None:
            raise ConventionError(convention=self, property="c33")
        return self._c33

    @c33.setter
    def c33(self, new_value: float):
        raise AttributeError(
            "It is intentionally forbidden to set properties of convention. "
            "Use correct methods of SpinHamiltonian class to change convention."
        )

    ################################################################################
    #                                  Four spins                                  #
    ################################################################################
    @property
    def c41(self) -> float:
        r"""
        Numerical factor before the (four spins & one site) sum of the Hamiltonian.
        """
        if self._c41 is None:
            raise ConventionError(convention=self, property="c41")
        return self._c41

    @c41.setter
    def c41(self, new_value: float):
        raise AttributeError(
            "It is intentionally forbidden to set properties of convention. "
            "Use correct methods of SpinHamiltonian class to change convention."
        )

    @property
    def c421(self) -> float:
        r"""
        Numerical factor before the (four spins & two sites (1+3)) sum of the Hamiltonian.
        """
        if self._c421 is None:
            raise ConventionError(convention=self, property="c421")
        return self._c421

    @c421.setter
    def c421(self, new_value: float):
        raise AttributeError(
            "It is intentionally forbidden to set properties of convention. "
            "Use correct methods of SpinHamiltonian class to change convention."
        )

    @property
    def c422(self) -> float:
        r"""
        Numerical factor before the (four spins & two sites (2+2)) sum of the Hamiltonian.
        """
        if self._c422 is None:
            raise ConventionError(convention=self, property="c422")
        return self._c422

    @c422.setter
    def c422(self, new_value: float):
        raise AttributeError(
            "It is intentionally forbidden to set properties of convention. "
            "Use correct methods of SpinHamiltonian class to change convention."
        )

    @property
    def c43(self) -> float:
        r"""
        Numerical factor before the (four spins & three sites) sum of the Hamiltonian.
        """
        if self._c43 is None:
            raise ConventionError(convention=self, property="c43")
        return self._c43

    @c43.setter
    def c43(self, new_value: float):
        raise AttributeError(
            "It is intentionally forbidden to set properties of convention. "
            "Use correct methods of SpinHamiltonian class to change convention."
        )

    @property
    def c44(self) -> float:
        r"""
        Numerical factor before the (four spins & four sites) sum of the Hamiltonian.
        """
        if self._c44 is None:
            raise ConventionError(convention=self, property="c44")
        return self._c44

    @c44.setter
    def c44(self, new_value: float):
        raise AttributeError(
            "It is intentionally forbidden to set properties of convention. "
            "Use correct methods of SpinHamiltonian class to change convention."
        )

    def __eq__(self, other):
        # Note semi-private attributes are compared intentionally, as
        # public ones will raise an error if not defined
        # If attributes are not defined in both conventions,
        # then that attribute is considered equal.
        return (
            self._multiple_counting == other._multiple_counting
            and self._spin_normalized == other._spin_normalized
            and self._c1 == other._c1
            and self._c21 == other._c21
            and self._c22 == other._c22
            and self._c31 == other._c31
            and self._c32 == other._c32
            and self._c33 == other._c33
            and self._c41 == other._c41
            and self._c421 == other._c421
            and self._c422 == other._c422
            and self._c43 == other._c43
            and self._c44 == other._c44
        )

    @staticmethod
    def get_predefined(name: str):
        r"""
        Returns one of the pre-defined conventions.

        Parameters
        ----------
        name : str
            Name of the desired pre-defined convention. Supported are

            * "tb2j"
            * "spinw"
            * vampire"

            Case-insensitive.

        Returns
        -------
        convention : :py:class:`.Convention`

        Examples
        --------

        .. doctest::

            >>> import magnopy
            >>> tb2j = magnopy.Convention.get_predefined("TB2J")
            >>> tb2j.summary()
            tb2j convention where
              * Bonds are counted multiple times in the sum;
              * Spin vectors are normalized to 1;
              * Undefined c1 factor;
              * c21 = -1.0;
              * c22 = -1.0;
              * Undefined c31 factor;
              * Undefined c32 factor;
              * Undefined c33 factor;
              * Undefined c41 factor;
              * Undefined c421 factor;
              * Undefined c422 factor;
              * Undefined c43 factor;
              * Undefined c44 factor.
            >>> spinW = magnopy.Convention.get_predefined("spinW")
            >>> spinW.summary()
            spinw convention where
              * Bonds are counted multiple times in the sum;
              * Spin vectors are not normalized;
              * Undefined c1 factor;
              * c21 = 1.0;
              * c22 = 1.0;
              * Undefined c31 factor;
              * Undefined c32 factor;
              * Undefined c33 factor;
              * Undefined c41 factor;
              * Undefined c421 factor;
              * Undefined c422 factor;
              * Undefined c43 factor;
              * Undefined c44 factor.
            >>> vampire = magnopy.Convention.get_predefined("Vampire")
            >>> vampire.summary()
            vampire convention where
              * Bonds are counted multiple times in the sum;
              * Spin vectors are normalized to 1;
              * Undefined c1 factor;
              * c21 = -1.0;
              * c22 = -0.5;
              * Undefined c31 factor;
              * Undefined c32 factor;
              * Undefined c33 factor;
              * Undefined c41 factor;
              * Undefined c421 factor;
              * Undefined c422 factor;
              * Undefined c43 factor;
              * Undefined c44 factor.
        """

        name = name.lower()

        if name not in _CONVENTIONS:
            ValueError(f"'{name}' convention is undefined.")

        return Convention(
            name=name,
            multiple_counting=_CONVENTIONS[name][0],
            spin_normalized=_CONVENTIONS[name][1],
            c21=_CONVENTIONS[name][2],
            c22=_CONVENTIONS[name][3],
            c31=_CONVENTIONS[name][4],
            c32=_CONVENTIONS[name][5],
            c33=_CONVENTIONS[name][6],
            c41=_CONVENTIONS[name][7],
            c421=_CONVENTIONS[name][8],
            c422=_CONVENTIONS[name][9],
            c43=_CONVENTIONS[name][10],
            c44=_CONVENTIONS[name][11],
        )

    def get_modified(
        self,
        multiple_counting: bool = None,
        spin_normalized: bool = None,
        c1: float = None,
        c21: float = None,
        c22: float = None,
        c31: float = None,
        c32: float = None,
        c33: float = None,
        c41: float = None,
        c421: float = None,
        c422: float = None,
        c43: float = None,
        c44: float = None,
        name: str = None,
    ):
        r"""
        Returns the new instance of the :py:class:`.Convention` class based on the called
        one with changed given properties.

        Parameters
        ----------
        multiple_counting : bool, optional
            Whether the pairs of spins are counted multiple times in the Hamiltonian's sums.
            Modified to the given value, if None, then kept the same as in the original convention.
        spin_normalized : bool, optional
            Whether spin vectors/operators are normalized to 1. If ``True``, then spin
            vectors/operators are normalized.
            Modified to the given value, if None, then kept the same as in the original convention.
        c1 : float, optional
            Numerical factor before the (one spin & one site) term of the Hamiltonian.
            Modified to the given value, if None, then kept the same as in the original convention.
        c21 : float, optional
            Numerical factor before the (two spins & one site) term of the Hamiltonian.
            Modified to the given value, if None, then kept the same as in the original convention.
        c22 : float, optional
            Numerical factor before the (two spins & two sites) term of the Hamiltonian.
            Modified to the given value, if None, then kept the same as in the original convention.
        c31 : float, optional
            Numerical factor before the (three spins & one site) term of the Hamiltonian.
            Modified to the given value, if None, then kept the same as in the original convention.
        c32 : float, optional
            Numerical factor before the (three spins & two sites) term of the Hamiltonian.
            Modified to the given value, if None, then kept the same as in the original convention.
        c33 : float, optional
            Numerical factor before the (three spins & three sites) term of the Hamiltonian.
            Modified to the given value, if None, then kept the same as in the original convention.
        c41 : float, optional
            Numerical factor before the (four spins & one site) term of the Hamiltonian.
            Modified to the given value, if None, then kept the same as in the original convention.
        c421 : float, optional
            Numerical factor before the (four spins & two sites & 1+3) term of the Hamiltonian.
            Modified to the given value, if None, then kept the same as in the original convention.
        c422 : float, optional
            Numerical factor before the (four spins & two sites & 2+2) term of the Hamiltonian.
            Modified to the given value, if None, then kept the same as in the original convention.
        c43 : float, optional
            Numerical factor before the (four spins & three sites) term of the Hamiltonian.
            Modified to the given value, if None, then kept the same as in the original convention.
        c44 : float, optional
            Numerical factor before the (four spins & four sites) term of the Hamiltonian.
            Modified to the given value, if None, then kept the same as in the original convention.
        name : str, optional
            A label for the convention. Any string, case-insensitive.
            Modified to the given value, if None, then kept the same as in the original convention.
        """

        if multiple_counting is None:
            multiple_counting = self._multiple_counting

        if spin_normalized is None:
            spin_normalized = self._spin_normalized

        if c1 is None:
            c1 = self._c1

        if c21 is None:
            c21 = self._c21

        if c22 is None:
            c22 = self._c22

        if c31 is None:
            c31 = self._c31

        if c32 is None:
            c32 = self._c32

        if c33 is None:
            c33 = self._c33

        if c41 is None:
            c41 = self._c41

        if c421 is None:
            c421 = self._c421

        if c422 is None:
            c422 = self._c422

        if c43 is None:
            c43 = self._c43

        if c44 is None:
            c44 = self._c44

        return Convention(
            spin_normalized=spin_normalized,
            multiple_counting=multiple_counting,
            c1=c1,
            c21=c21,
            c22=c22,
            c31=c31,
            c32=c32,
            c33=c33,
            c41=c41,
            c421=c421,
            c422=c422,
            c43=c43,
            c44=c44,
        )


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from copy import deepcopy

import numpy as np
from wulfric import add_sugar
from wulfric.crystal import get_distance

from magnopy._spinham._c1 import _add_1, _p1, _remove_1
from magnopy._spinham._c21 import _add_21, _p21, _remove_21
from magnopy._spinham._c22 import _add_22, _p22, _remove_22
from magnopy._spinham._c31 import _add_31, _p31, _remove_31
from magnopy._spinham._c32 import _add_32, _p32, _remove_32
from magnopy._spinham._c33 import _add_33, _p33, _remove_33
from magnopy._spinham._c41 import _add_41, _p41, _remove_41
from magnopy._spinham._c43 import _add_43, _p43, _remove_43
from magnopy._spinham._c44 import _add_44, _p44, _remove_44
from magnopy._spinham._c421 import _add_421, _p421, _remove_421
from magnopy._spinham._c422 import _add_422, _p422, _remove_422
from magnopy._spinham._convention import Convention
from magnopy._spinham._validators import _validate_atom_index, _validate_unit_cell_index

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


class SpinHamiltonian:
    r"""
    Spin Hamiltonian.

    Parameters
    ----------
    convention : :py:class:`.Convention` or str
        A convention of the spin Hamiltonian.
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.
    atoms : dict
        Dictionary with atoms.

    """

    def __init__(self, cell, atoms, convention) -> None:
        self._cell = np.array(cell)

        self._atoms = add_sugar(atoms)

        # Only the magnetic sites
        self._magnetic_atoms = None
        self._index_map = None

        self._convention = convention

        # [[alpha, parameter], ...]
        self._1 = []

        # [[alpha, parameter], ...]
        self._21 = []

        # [[alpha, beta, nu, parameter], ...]
        self._22 = []

        # [[alpha, parameter], ...]
        self._31 = []

        # [[alpha, beta, nu, parameter], ...]
        self._32 = []

        # [[alpha, beta, gamma, nu, lambda, parameter], ...]
        self._33 = []

        # [[alpha, parameter], ...]
        self._41 = []

        # [[alpha, beta, nu, parameter], ...]
        self._421 = []

        # [[alpha, beta, nu, parameter], ...]
        self._422 = []

        # [[alpha, beta, gamma, nu, lambda, parameter], ...]
        self._43 = []

        # [[alpha, beta, gamma, epsilon, nu, lambda, rho, parameter], ...]
        self._44 = []

    ############################################################################
    #                              Cell and Atoms                              #
    ############################################################################

    @property
    def cell(self):
        r"""
        Cell of the crystal on which the Hamiltonian is build.

        Returns
        -------
        cell : (3, 3) :numpy:`ndarray`
            Matrix of the cell, rows are lattice vectors.

        Notes
        -----
        If is not recommended to change the ``cell`` property after the creation of
        :py:class:`.SpinHamiltonian`. In fact an attempt to do so will raise an
        ``AttributeError``:

        .. doctest::

            >>> import numpy as np
            >>> import magnopy
            >>> convention = magnopy.Convention()
            >>> spinham = magnopy.SpinHamiltonian(cell = np.eye(3), atoms={}, convention=convention)
            >>> spinham.cell = 2 * np.eye(3)
            Traceback (most recent call last):
            ...
            AttributeError: Change of the cell attribute is not supported after the creation of SpinHamiltonian instance. If you need to modify cell, then use pre-defined methods of SpinHamiltonian or create a new one.

        Use pre-defined methods of the :py:class:`.SpinHamiltonian` class to safely
        modify the cell.

        If you need to change the cell attribute, then use

        .. doctest::

            >>> import numpy as np
            >>> import magnopy
            >>> convention = magnopy.Convention()
            >>> spinham = magnopy.SpinHamiltonian(cell = np.eye(3), atoms={}, convention=convention)
            >>> spinham.cell
            array([[1., 0., 0.],
                   [0., 1., 0.],
                   [0., 0., 1.]])
            >>> spinham._cell = 2 * np.eye(3)
            >>> spinham.cell
            array([[2., 0., 0.],
                   [0., 2., 0.],
                   [0., 0., 2.]])


        In the latter case correct behavior of magnopy **is not** guaranteed. Use only
        if you have a deep understanding of the magnopy source code.
        """

        return self._cell

    @cell.setter
    def cell(self, new_value):
        raise AttributeError(
            f"Change of the cell attribute is not supported after "
            "the creation of SpinHamiltonian instance. If you need to modify cell, "
            "then use pre-defined methods of SpinHamiltonian or create a new one."
        )

    @property
    def atoms(self) -> dict:
        r"""
        Atoms of the crystal on which the Hamiltonian is build.

        Returns
        -------
        atoms : dict (with added sugar)
            Dictionary with the atoms.

        Notes
        -----
        If is not recommended to change the atoms property after the creation of
        :py:class:`.SpinHamiltonian`. In fact an attempt to do so will raise an
        ``AttributeError``:

        .. doctest::

            >>> import numpy as np
            >>> import magnopy
            >>> convention = magnopy.Convention()
            >>> spinham = magnopy.SpinHamiltonian(cell = np.eye(3), atoms={}, convention=convention)
            >>> spinham.atoms = {"names" : ["Cr"]}
            Traceback (most recent call last):
            ...
            AttributeError: Change of the atoms dictionary is not supported after the creation of SpinHamiltonian instance. If you need to modify atoms, then use pre-defined methods of SpinHamiltonian or create a new one.

        Use pre-defined methods of the :py:class:`.SpinHamiltonian` class to safely
        modify atoms.

        If you need to change the whole dictionary at once, then use

        .. doctest::

            >>> import numpy as np
            >>> import magnopy
            >>> convention = magnopy.Convention()
            >>> spinham = magnopy.SpinHamiltonian(cell = np.eye(3), atoms={}, convention=convention)
            >>> spinham.atoms
            {}
            >>> spinham._atoms = {"names" : ["Cr"]}
            >>> spinham.atoms
            {'names': ['Cr']}

        In the latter case correct behavior of magnopy **is not** guaranteed. Use only
        if you have a deep understanding of the magnopy source code.
        """

        return self._atoms

    @atoms.setter
    def atoms(self, new_value):
        raise AttributeError(
            f"Change of the atoms dictionary is not supported after "
            "the creation of SpinHamiltonian instance. If you need to modify atoms, "
            "then use pre-defined methods of SpinHamiltonian or create a new one."
        )

    def _reset_internals(self):
        self._index_map = None
        self._magnetic_atoms = None

    def _update_internals(self):
        # Identify magnetic sites
        indices = set()

        for alpha, _ in self._1:
            indices.add(alpha)

        for alpha, _ in self._21:
            indices.add(alpha)

        for alpha, beta, _, _ in self._22:
            indices.add(alpha)
            indices.add(beta)

        for alpha, _ in self._31:
            indices.add(alpha)

        for alpha, beta, _, _ in self._32:
            indices.add(alpha)
            indices.add(beta)

        for alpha, beta, gamma, _, _, _ in self._33:
            indices.add(alpha)
            indices.add(beta)
            indices.add(gamma)

        for alpha, _ in self._41:
            indices.add(alpha)

        for alpha, beta, _, _ in self._421:
            indices.add(alpha)
            indices.add(beta)

        for alpha, beta, _, _ in self._422:
            indices.add(alpha)
            indices.add(beta)

        for alpha, beta, gamma, _, _, _ in self._43:
            indices.add(alpha)
            indices.add(beta)
            indices.add(gamma)

        for alpha, beta, gamma, epsilon, _, _, _, _ in self._44:
            indices.add(alpha)
            indices.add(beta)
            indices.add(gamma)
            indices.add(epsilon)

        indices = sorted(list(indices))

        # Create index map
        self._index_map = [None for _ in range(len(self.atoms.names))]
        for i in range(len(indices)):
            self._index_map[indices[i]] = i

        # Create magnetic atoms dictionary
        self._magnetic_atoms = add_sugar({})

        for key in self.atoms:
            self._magnetic_atoms[key] = []

            for full_index in indices:
                self._magnetic_atoms[key].append(self.atoms[key][full_index])

    @property
    def index_map(self):
        r"""
        Index map from all atoms to the magnetic ones.

        Returns
        -------
        index_map (L, ) list of int
            Index map. Integers. ``L = len(spinham.atoms.names)``
        """

        if self._index_map is None:
            self._update_internals()

        return self._index_map

    @property
    def magnetic_atoms(self):
        r"""
        Magnetic atoms of the spin Hamiltonian.

        Magnetic atom is defined as an atom with at least one parameter associated with
        it.

        This property is dynamically computed at every call.

        Returns
        -------
        magnetic_atoms : list of int
            Indices of magnetic atoms in the ``spinham.atoms``. Sorted.

        See Also
        --------
        M
        """

        if self._magnetic_atoms is None:
            self._update_internals()

        return self._magnetic_atoms

    @property
    def M(self):
        r"""
        Number of spins (magnetic atoms) in the unit cell.

        Returns
        -------
        M : int
            Number of spins (magnetic atoms) in the unit cell.

        See Also
        --------
        magnetic_atoms
        """

        return len(self.magnetic_atoms.names)

    ############################################################################
    #                          Convention properties                           #
    ############################################################################
    @property
    def convention(self) -> Convention:
        r"""
        Convention of the spin Hamiltonian.

        See Also
        --------
        Convention
        """

        return self._convention

    @convention.setter
    def convention(self, new_convention: Convention):
        self._set_multiple_counting(new_convention._multiple_counting)

        self._set_spin_normalization(new_convention._spin_normalized)

        self._set_c1(new_convention._c1)

        self._set_c21(new_convention._c21)
        self._set_c22(new_convention._c22)

        self._set_c31(new_convention._c31)
        self._set_c32(new_convention._c32)
        self._set_c33(new_convention._c33)

        self._set_c41(new_convention._c41)
        self._set_c421(new_convention._c421)
        self._set_c422(new_convention._c422)
        self._set_c43(new_convention._c43)
        self._set_c44(new_convention._c44)

        self._convention = new_convention

    def _set_multiple_counting(self, multiple_counting: bool) -> None:
        if multiple_counting is None or self.convention._multiple_counting is None:
            return

        multiple_counting = bool(multiple_counting)

        if self.convention.multiple_counting == multiple_counting:
            return

        # For (two spins & two sites)

        # It was absent before
        if multiple_counting:
            factor = 0.5
        # It was present before
        else:
            factor = 2.0

        for index in range(len(self._22)):
            self._22[index][3] = self._22[index][3] * factor

        # For (three spins & two sites)

        # It was absent before
        if multiple_counting:
            factor = 0.5
        # It was present before
        else:
            factor = 2.0

        for index in range(len(self._32)):
            self._32[index][3] = self._32[index][3] * factor

        # For (three spins & three sites)

        # It was absent before
        if multiple_counting:
            factor = 1 / 6
        # It was present before
        else:
            factor = 6

        for index in range(len(self._33)):
            self._33[index][5] = self._33[index][5] * factor

        # For (four spins & two sites (3+1))

        # It was absent before
        if multiple_counting:
            factor = 0.5
        # It was present before
        else:
            factor = 2.0

        for index in range(len(self._421)):
            self._421[index][3] = self._421[index][3] * factor

        # For (four spins & two sites (2+2))

        # It was absent before
        if multiple_counting:
            factor = 0.5
        # It was present before
        else:
            factor = 2.0

        for index in range(len(self._422)):
            self._422[index][3] = self._422[index][3] * factor

        # For (four spins & three sites)

        # It was absent before
        if multiple_counting:
            factor = 1 / 6
        # It was present before
        else:
            factor = 6

        for index in range(len(self._43)):
            self._43[index][5] = self._43[index][5] * factor

        # For (four spins & four sites)

        # It was absent before
        if multiple_counting:
            factor = 1 / 24
        # It was present before
        else:
            factor = 24

        for index in range(len(self._44)):
            self._44[index][7] = self._44[index][7] * factor

    def _set_spin_normalization(self, spin_normalized: bool) -> None:
        if spin_normalized is None or self.convention._spin_normalized is None:
            return

        spin_normalized = bool(spin_normalized)

        if self.convention.spin_normalized == spin_normalized:
            return

        # Before it was not normalized
        if spin_normalized:
            # For (one spin & one site)
            for index in range(len(self._1)):
                alpha = self._1[index][0]
                self._1[index][1] = self._1[index][1] * self.atoms.spins[alpha]
            # For (two spins & one site)
            for index in range(len(self._21)):
                alpha = self._21[index][0]
                self._21[index][1] = self._21[index][1] * self.atoms.spins[alpha] ** 2
            # For (two spins & two sites)
            for index in range(len(self._22)):
                alpha = self._22[index][0]
                beta = self._22[index][1]
                self._22[index][3] = self._22[index][3] * (
                    self.atoms.spins[alpha] * self.atoms.spins[beta]
                )
            # For (three spins & one site)
            for index in range(len(self._31)):
                alpha = self._31[index][0]
                self._31[index][1] = self._31[index][1] * self.atoms.spins[alpha] ** 3
            # For (three spins & two sites)
            for index in range(len(self._32)):
                alpha = self._32[index][0]
                beta = self._32[index][1]
                self._32[index][3] = self._32[index][3] * (
                    self.atoms.spins[alpha] ** 2 * self.atoms.spins[beta]
                )
            # For (three spins & three sites)
            for index in range(len(self._33)):
                alpha = self._33[index][0]
                beta = self._33[index][1]
                gamma = self._33[index][2]
                self._33[index][5] = self._33[index][5] * (
                    self.atoms.spins[alpha]
                    * self.atoms.spins[beta]
                    * self.atoms.spins[gamma]
                )
            # For (four spins & one site)
            for index in range(len(self._41)):
                alpha = self._41[index][0]
                self._41[index][1] = self._41[index][1] * self.atoms.spins[alpha] ** 4
            # For (four spins & two sites (3+1))
            for index in range(len(self._421)):
                alpha = self._421[index][0]
                beta = self._421[index][1]
                self._421[index][3] = self._421[index][3] * (
                    self.atoms.spins[alpha] ** 3 * self.atoms.spins[beta]
                )
            # For (four spins & two sites (2+2))
            for index in range(len(self._422)):
                alpha = self._422[index][0]
                beta = self._422[index][1]
                self._422[index][3] = self._422[index][3] * (
                    self.atoms.spins[alpha] ** 2 * self.atoms.spins[beta] ** 2
                )
            # For (four spins & three sites)
            for index in range(len(self._43)):
                alpha = self._43[index][0]
                beta = self._43[index][1]
                gamma = self._43[index][2]
                self._43[index][5] = self._43[index][5] * (
                    self.atoms.spins[alpha] ** 2
                    * self.atoms.spins[beta]
                    * self.atoms.spins[gamma]
                )
            # For (four spins & four sites)
            for index in range(len(self._44)):
                alpha = self._44[index][0]
                beta = self._44[index][1]
                gamma = self._44[index][2]
                epsilon = self._44[index][3]
                self._44[index][7] = self._44[index][7] * (
                    self.atoms.spins[alpha]
                    * self.atoms.spins[beta]
                    * self.atoms.spins[gamma]
                    * self.atoms.spins[epsilon]
                )
        # Before it was normalized
        else:
            # For (one spin & one site)
            for index in range(len(self._1)):
                alpha = self._1[index][0]
                self._1[index][1] = self._1[index][1] / self.atoms.spins[alpha]
            # For (two spins & one site)
            for index in range(len(self._21)):
                alpha = self._21[index][0]
                self._21[index][1] = self._21[index][1] / self.atoms.spins[alpha] ** 2
            # For (two spins & two sites)
            for index in range(len(self._22)):
                alpha = self._22[index][0]
                beta = self._22[index][1]
                self._22[index][3] = self._22[index][3] / (
                    self.atoms.spins[alpha] * self.atoms.spins[beta]
                )
            # For (three spins & one site)
            for index in range(len(self._31)):
                alpha = self._31[index][0]
                self._31[index][1] = self._31[index][1] / self.atoms.spins[alpha] ** 3
            # For (three spins & two sites)
            for index in range(len(self._32)):
                alpha = self._32[index][0]
                beta = self._32[index][1]
                self._32[index][3] = self._32[index][3] / (
                    self.atoms.spins[alpha] ** 2 * self.atoms.spins[beta]
                )
            # For (three spins & three sites)
            for index in range(len(self._33)):
                alpha = self._33[index][0]
                beta = self._33[index][1]
                gamma = self._33[index][2]
                self._33[index][5] = self._33[index][5] / (
                    self.atoms.spins[alpha]
                    * self.atoms.spins[beta]
                    * self.atoms.spins[gamma]
                )
            # For (four spins & one site)
            for index in range(len(self._41)):
                alpha = self._41[index][0]
                self._41[index][1] = self._41[index][1] / self.atoms.spins[alpha] ** 4
            # For (four spins & two sites (3+1))
            for index in range(len(self._421)):
                alpha = self._421[index][0]
                beta = self._421[index][1]
                self._421[index][3] = self._421[index][3] / (
                    self.atoms.spins[alpha] ** 3 * self.atoms.spins[beta]
                )
            # For (four spins & two sites (2+2))
            for index in range(len(self._422)):
                alpha = self._422[index][0]
                beta = self._422[index][1]
                self._422[index][3] = self._422[index][3] / (
                    self.atoms.spins[alpha] ** 2 * self.atoms.spins[beta] ** 2
                )
            # For (four spins & three sites)
            for index in range(len(self._43)):
                alpha = self._43[index][0]
                beta = self._43[index][1]
                gamma = self._43[index][2]
                self._43[index][5] = self._43[index][5] / (
                    self.atoms.spins[alpha] ** 2
                    * self.atoms.spins[beta]
                    * self.atoms.spins[gamma]
                )
            # For (four spins & four sites)
            for index in range(len(self._44)):
                alpha = self._44[index][0]
                beta = self._44[index][1]
                gamma = self._44[index][2]
                epsilon = self._44[index][3]
                self._44[index][7] = self._44[index][7] / (
                    self.atoms.spins[alpha]
                    * self.atoms.spins[beta]
                    * self.atoms.spins[gamma]
                    * self.atoms.spins[epsilon]
                )

    def _set_c1(self, new_c1: float) -> None:
        if new_c1 is None or self.convention._c1 is None:
            return

        new_c1 = float(new_c1)

        if self.convention.c1 == new_c1:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._1)):
            self._1[index][1] = self._1[index][1] * self.convention.c1 / new_c1

    def _set_c21(self, new_c21: float) -> None:
        if new_c21 is None or self.convention._c21 is None:
            return

        new_c21 = float(new_c21)

        if self.convention.c21 == new_c21:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._21)):
            self._21[index][1] = self._21[index][1] * self.convention.c21 / new_c21

    def _set_c22(self, new_c22: float) -> None:
        if new_c22 is None or self.convention._c22 is None:
            return

        new_c22 = float(new_c22)

        if self.convention.c22 == new_c22:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._22)):
            self._22[index][3] = self._22[index][3] * self.convention.c22 / new_c22

    def _set_c31(self, new_c31: float) -> None:
        if new_c31 is None or self.convention._c31 is None:
            return

        new_c31 = float(new_c31)

        if self.convention.c31 == new_c31:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._31)):
            self._31[index][1] = self._31[index][1] * self.convention.c31 / new_c31

    def _set_c32(self, new_c32: float) -> None:
        if new_c32 is None or self.convention._c32 is None:
            return

        new_c32 = float(new_c32)

        if self.convention.c32 == new_c32:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._32)):
            self._32[index][3] = self._32[index][3] * self.convention.c32 / new_c32

    def _set_c33(self, new_c33: float) -> None:
        if new_c33 is None or self.convention._c33 is None:
            return

        new_c33 = float(new_c33)

        if self.convention.c33 == new_c33:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._33)):
            self._33[index][5] = self._33[index][5] * self.convention.c33 / new_c33

    def _set_c41(self, new_c41: float) -> None:
        if new_c41 is None or self.convention._c41 is None:
            return

        new_c41 = float(new_c41)

        if self.convention.c41 == new_c41:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._41)):
            self._41[index][1] = self._41[index][1] * self.convention.c41 / new_c41

    def _set_c421(self, new_c421: float) -> None:
        if new_c421 is None or self.convention._c421 is None:
            return

        new_c421 = float(new_c421)

        if self.convention.c421 == new_c421:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._421)):
            self._421[index][3] = self._421[index][3] * self.convention.c421 / new_c421

    def _set_c422(self, new_c422: float) -> None:
        if new_c422 is None or self.convention._c422 is None:
            return

        new_c422 = float(new_c422)

        if self.convention.c422 == new_c422:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._422)):
            self._422[index][3] = self._422[index][3] * self.convention.c422 / new_c422

    def _set_c43(self, new_c43: float) -> None:
        if new_c43 is None or self.convention._c43 is None:
            return

        new_c43 = float(new_c43)

        if self.convention.c43 == new_c43:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._43)):
            self._43[index][5] = self._43[index][5] * self.convention.c43 / new_c43

    def _set_c44(self, new_c44: float) -> None:
        if new_c44 is None or self.convention._c44 is None:
            return

        new_c44 = float(new_c44)

        if self.convention.c44 == new_c44:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._44)):
            self._44[index][7] = self._44[index][7] * self.convention.c44 / new_c44

    def add_magnetic_field(self, h) -> None:
        r"""
        Adds external magnetic field to the Hamiltonian in the form of one spin
        parameters.

        .. math::

            \mu_B \boldsymbol{h}\cdot\sum_{\mu,\alpha} g_{\alpha} \boldsymbol{S}_{\mu,\alpha}

        Parameters
        ----------
        h : (3, ) |array-like|_
            Vector of magnetic field given in the units of Tesla.

        Notes
        -----
        If Hamiltonian read from TB2J or GROGU, then by default g factors are positive (=2).
        Therefore, to minimize the energy the magnetic moment will be aligned with the
        direction of the external field. But pin vector will be directed opposite to the
        direction of the magnetic field.

        Magnetic field is added *only* to the magnetic atoms. In other words, magnetic
        field is added only to the atoms that have at least one other parameter
        associated with them.
        """

        if self.convention._c1 is None:
            self.convention._c1 = 1.0

        h = np.array(h, dtype=float)

        BOHR_MAGNETON = 0.057883818060  # meV / Tesla

        new_1 = []

        for i in range(0, self.M):
            new_1.append(
                BOHR_MAGNETON
                * self.magnetic_atoms.g_factors[i]
                * h
                / self.convention.c1
            )

        for alpha, parameter in self._1:
            new_1[self.index_map[alpha]] = new_1[self.index_map[alpha]] + parameter

        self._1 = list(map(list, zip(range(self.M), new_1)))

    ############################################################################
    #                                Copy getter                               #
    ############################################################################
    def copy(self):
        R"""
        Returns a new, independent copy of the same Hamiltonian.

        Returns
        -------
        spinham : :py:class:`.SpinHamiltonian`
            A new instance of the same Hamiltonian.
        """

        return deepcopy(self)

    ############################################################################
    #                            One spin & one site                           #
    ############################################################################
    p1 = _p1
    add_1 = _add_1
    remove_1 = _remove_1

    ############################################################################
    #                           Two spins & one site                           #
    ############################################################################
    p21 = _p21
    add_21 = _add_21
    remove_21 = _remove_21

    ############################################################################
    #                           Two spins & two sites                          #
    ############################################################################
    p22 = _p22
    add_22 = _add_22
    remove_22 = _remove_22

    ############################################################################
    #                          Three spins & one site                          #
    ############################################################################
    p31 = _p31
    add_31 = _add_31
    remove_31 = _remove_31

    ############################################################################
    #                          Three spins & two sites                         #
    ############################################################################
    p32 = _p32
    add_32 = _add_32
    remove_32 = _remove_32

    ############################################################################
    #                         Three spins & three sites                        #
    ############################################################################
    p33 = _p33
    add_33 = _add_33
    remove_33 = _remove_33

    ############################################################################
    #                           Four spins & one site                          #
    ############################################################################
    p41 = _p41
    add_41 = _add_41
    remove_41 = _remove_41

    ############################################################################
    #                          Four spins & two sites (3+1)                    #
    ############################################################################
    p421 = _p421
    add_421 = _add_421
    remove_421 = _remove_421

    ############################################################################
    #                          Four spins & two sites (2+2)                    #
    ############################################################################
    p422 = _p422
    add_422 = _add_422
    remove_422 = _remove_422

    ############################################################################
    #                         Four spins & three sites                         #
    ############################################################################
    p43 = _p43
    add_43 = _add_43
    remove_43 = _remove_43

    ############################################################################
    #                          Four spins & four sites                         #
    ############################################################################
    p44 = _p44
    add_44 = _add_44
    remove_44 = _remove_44


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from functools import cmp_to_key

import numpy as np
from numpy.linalg import LinAlgError

from magnopy._exceptions import ColpaFailed

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def _check_grand_dynamical_matrix(D):
    r"""
    Check that grand dynamical matrix is (2N, 2N) matrix

    Parameters
    ----------
    D : |array-like|_
        Candidate for the grand dynamical matrix

    Returns
    -------
    D : (2N, 2N) :numpy:`ndarray`
        Grand dynamical matrix.
    N : int

    Raises
    ------
    ValueError
        If the check is not passed
    """

    D = np.array(D)

    if len(D.shape) != 2:
        raise ValueError(f"Grand dynamical matrix is not 2-dimensional, got {D.shape}.")

    if D.shape[0] != D.shape[1]:
        raise (f"Grand dynamical matrix is not square, got {D.shape}.")

    if D.shape[0] % 2 != 0:
        raise (f"Size of the grand dynamical matrix is not even, got {D.shape}.")

    return D, D.shape[0] // 2


def _inverse_by_colpa(matrix):
    # Compute G from G^-1 following Colpa, see equation (3.7) for details
    matrix = np.conjugate(matrix).T
    matrix[:N, N:] *= -1
    matrix[N:, :N] *= -1

    return matrix


def solve_via_colpa(D, return_inverse=False):
    r"""
    Diagonalize grand-dynamical matrix following the method of Colpa [1]_.

    Algorithm is described in section 3, remark 1 of [1]_.

    Solves the Bogoliubov Hamiltonian of the form:

    .. math::

        \hat{H} = \sum_{r^{\prime}, r = 1}^m
        a_{r^{\prime}}^{\dagger}\boldsymbol{\Delta}_1^{r^{\prime}r}a_r +
        a_{r^{\prime}}^{\dagger}\boldsymbol{\Delta}_2^{r^{\prime}r}a_{m+r}^{\dagger} +
        a_{m+r^{\prime}}^{\dagger}\boldsymbol{\Delta}_3^{r^{\prime}r}a_r +
        a_{m+r^{\prime}}^{\dagger}\boldsymbol{\Delta}_4^{r^{\prime}r}a_{m+r}^{\dagger}

    In a matrix form the Hamiltonian is:

    .. math::

        \hat{H} = \boldsymbol{\cal A}^{\dagger} \boldsymbol{D} \boldsymbol{\cal A}

    where

    .. math::

        \boldsymbol{\cal A} =
        \begin{pmatrix}
            a_1 \\
            \cdots \\
            a_m \\
            a_{m+1} \\
            \cdots \\
            a_{2m} \\
        \end{pmatrix}

    After diagonalization the Hamiltonian is:

    .. math::

        \hat{H}
        =
        \boldsymbol{\cal B}^{\dagger} \boldsymbol{\mathcal{E}} \boldsymbol{\cal B}

    Parameters
    ----------
    D : (2N, 2N) |array-like|_
        Grand dynamical matrix. If it is Hermitian and positive-defined, then obtained
        eigenvalues are positive and real.

        .. math::

            \boldsymbol{\mathcal{D}} = \begin{pmatrix}
                \boldsymbol{\Delta_1} & \boldsymbol{\Delta_2} \\
                \boldsymbol{\Delta_3} & \boldsymbol{\Delta_4}
            \end{pmatrix}
    return_inverse : bool, default False
        Whether to return :math:`(\boldsymbol{G}^{\dagger})^{-1}` instead of
        :math:`\boldsymbol{G}^{\dagger}`.

    Returns
    -------
    E : (2N,) :numpy:`ndarray`
        The eigenvalues.
        It is an array of the diagonal elements of the
        diagonal matrix :math:`\boldsymbol{\mathcal{E}}` of the diagonalized Hamiltonian.
        First N elements correspond to the
        :math:`b^{\dagger}(\boldsymbol{k})b(\boldsymbol{k})` and last N elements - to
        the :math:`b(-\boldsymbol{k})b^{\dagger}(-\boldsymbol{k})`.

        Eigenvalues are sorted individually for the first N and the last N elements,
        based on the transformation matrix and not on the values of E itself.

        It is a diagonal of the matrix

        .. math::

            \boldsymbol{\mathcal{E}}
            =
            (\boldsymbol{G}^{\dagger})^{-1} \boldsymbol{D} \boldsymbol{G}^{-1}

    G : (2N, 2N) :numpy:`ndarray`
        Transformation matrix, which changes the basis from the original set of bosonic
        operators :math:`\boldsymbol{a}` to the set of new bosonic operators
        :math:`\boldsymbol{b}` which diagonalize the Hamiltonian:

        .. math::

            \boldsymbol{\cal A} = \boldsymbol{G} \boldsymbol{\cal B}

            \boldsymbol{\cal B} = \boldsymbol{G}^{-1} \boldsymbol{\cal A}

        The rows are ordered in the same way as the eigenvalues.

        If ``return_inverse == False``, then :math:`\boldsymbol{G}^{\dagger}` is
        returned.

        If ``return_inverse == True``, then :math:`(\boldsymbol{G}^{\dagger})^{-1}` is
        returned.

    Raises
    ------
    ColpaFailed
        If the algorithm fails. Typically it means that the grand dynamical matrix
        :math:`\boldsymbol{D}` is not positive-defined.
    ValueError
        If the grand dynamical matrix is not square or its shape is not even.

    References
    ----------
    .. [1] Colpa, J.H.P., 1978.
        Diagonalization of the quadratic boson hamiltonian.
        Physica A: Statistical Mechanics and its Applications,
        93(3-4), pp.327-353.
    """

    D, N = _check_grand_dynamical_matrix(D)

    g = np.diag(np.concatenate((np.ones(N), -np.ones(N))))

    try:
        # In Colpa article decomposition is K^{\dag}K, while numpy gives KK^{\dag}
        K = np.conjugate(np.linalg.cholesky(D)).T
    except LinAlgError:
        raise ColpaFailed

    L, U = np.linalg.eig(K @ g @ np.conjugate(K).T)

    # Sort with respect to L, in descending order
    U = np.concatenate((L[:, None], U.T), axis=1).T
    # U = np.concatenate((L[np.newaxis, :], U), axis=0)
    U = U[:, np.argsort(U[0])]
    L = U[0, ::-1]
    U = U[1:, ::-1]
    E = g @ L

    G_minus_one = np.linalg.inv(K) @ U @ np.sqrt(np.diag(E))

    # Sort first N and second N individually based on the transformation matrix
    tmp = np.concatenate((E[:, np.newaxis], G_minus_one), axis=1)

    def compare(array1, array2):
        difference = np.round(array1 - array2, decimals=15)
        if np.allclose(difference, np.zeros(difference.shape)):
            return 0.0

        return difference[np.nonzero(difference)[0][0]]

    upper_part = np.array(sorted(tmp[:N], key=cmp_to_key(compare)))
    lower_part = np.array(sorted(tmp[N:], key=cmp_to_key(compare)))

    E = np.concatenate((upper_part[:, 0], lower_part[:, 0]))
    G_minus_one = np.concatenate((upper_part[:, 1:], lower_part[:, 1:]), axis=0)

    if return_inverse:
        return E, G_minus_one

    return E, _inverse_by_colpa(G_minus_one)


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


R"""
Functions for the common manipulation with the exchange parameters.
"""

import numpy as np

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def get_matrix_parameter(iso=None, aniso=None, dmi=None) -> np.ndarray:
    r"""
    Computes :math:`3 \times 3` matrix from different parts of it.

    Parameters
    ----------
    iso : int or float, optional
        Value of isotropic exchange parameter.
    aniso : (3, 3) |array-like|_, optional
        Traceless, symmetric 3 x 3 matrix of exchange anisotropy.
        Note: Input matrix is symmetrized and made traceless.
    dmi : (3,) |array-like|_, optional
        Dzyaroshinsky-Moria interaction vector :math:`(D_x, D_y, D_z)`.

    Returns
    -------
    matrix : (3, 3) :numpy:`ndarray`
        Full matrix of exchange parameter. If no input provided, then returns matrix of
        zeros.

    Notes
    -----
    Full matrix of the exchange parameter can be decomposed into three common parts:
    isotropic (``iso``), traceless symmetric anisotropy
    (``aniso``) and antisymmetric anisotropy (``dmi``)

    .. math::

        \boldsymbol{J}
        =
        \begin{pmatrix}
            J_{iso} & 0 & 0 \\
            0 & J_{iso} & 0 \\
            0 & 0 & J_{iso} \\
        \end{pmatrix}
        +
        \begin{pmatrix}
            S^{xx} & S^{xy} & S^{xz} \\
            S^{xy} & S^{yy} & S^{yz} \\
            S^{xz} & S^{yz} & S^{zz} \\
        \end{pmatrix}
        +
        \begin{pmatrix}
            0 & D^z & -D^y \\
            -D^z & 0 & D^x \\
            D^y & -D^x & 0 \\
        \end{pmatrix}

    Examples
    --------

    .. doctest::

        >>> import magnopy
        >>> magnopy._spinham.get_matrix_parameter(iso=1)
        array([[1., 0., 0.],
               [0., 1., 0.],
               [0., 0., 1.]])
        >>> magnopy._spinham.get_matrix_parameter(dmi = (1, 2, 0))
        array([[ 0.,  0., -2.],
               [ 0.,  0.,  1.],
               [ 2., -1.,  0.]])

    """

    matrix = np.zeros((3, 3), dtype=float)

    if iso is not None:
        matrix += iso * np.eye(3, dtype=float)

    if aniso is not None:
        aniso = np.array(aniso, dtype=float)

        # Make sure that it is traceless
        aniso -= np.eye(3) * np.linalg.trace(aniso) / 3

        # Make sure that is is symmetric
        aniso = (aniso + aniso.T) / 2

        matrix += aniso

    if dmi is not None:
        matrix += [
            [0, dmi[2], -dmi[1]],
            [-dmi[2], 0, dmi[0]],
            [dmi[1], -dmi[0], 0],
        ]

    return matrix


def get_isotropic_parameter(matrix, matrix_form=False):
    r"""
    Computes isotropic parameter from full matrix parameter.

    Parameters
    ----------
    matrix : (3, 3) |array-like|_
        Full matrix of the parameter.
    matrix_form : bool, default False
        Whether to return isotropic part of the matrix instead of isotropic parameter.

    Returns
    -------
    iso : float or (3, 3) :numpy:`ndarray`
        Isotropic parameter. If ``matrix_form == True``, then return a matrix.

    Notes
    -----

    Isotropic parameter is defined as

    .. math::

        J_{iso} = \dfrac{tr(\boldsymbol{J})}{3}

    In a matrix form it is

    .. math::

        \begin{pmatrix}
            J_{iso} & 0 & 0 \\
            0 & J_{iso} & 0 \\
            0 & 0 & J_{iso} \\
        \end{pmatrix}

    Examples
    --------

    .. doctest::

        >>> import magnopy
        >>> matrix = [[1, 3, 4], [-1, -2, 3], [4, 0, 10]]
        >>> magnopy._spinham.get_isotropic_parameter(matrix)
        3.0
        >>> magnopy._spinham.get_isotropic_parameter(matrix, matrix_form=True)
        array([[3., 0., 0.],
               [0., 3., 0.],
               [0., 0., 3.]])
    """

    iso = np.linalg.trace(matrix) / 3

    if matrix_form:
        return iso * np.eye(3, dtype=float)

    return float(iso)


def get_anisotropic_parameter(matrix):
    r"""
    Computes traceless, symmetric anisotropy from full matrix parameter.

    Parameters
    ----------
    matrix : (3, 3) |array-like|_
        Full matrix of the parameter.

    Returns
    -------
    aniso : float or (3, 3) :numpy:`ndarray`
        Matrix of a traceless, symmetric anisotropy.

    Notes
    -----

    Traceless, symmetric anisotropy is defined as

    .. math::

        \boldsymbol{J}_{S}
        =
        \dfrac{\boldsymbol{J} + \boldsymbol{J}.T}{2}
        -
        J_{iso} \cdot \boldsymbol{I}

    Examples
    --------

    .. doctest::

        >>> import magnopy
        >>> matrix = [[1, 3, 4], [-1, -2, 0], [4, 0, 10]]
        >>> magnopy._spinham.get_anisotropic_parameter(matrix)
        array([[-2.,  1.,  4.],
               [ 1., -5.,  0.],
               [ 4.,  0.,  7.]])
    """

    matrix = np.array(matrix)

    return (matrix + matrix.T) / 2 - get_isotropic_parameter(
        matrix=matrix, matrix_form=True
    )


def get_dmi(matrix, matrix_form=False):
    r"""
    Computes DMI parameter from full matrix parameter.

    Parameters
    ----------
    matrix : (3, 3) |array-like|_
        Full matrix of the parameter.
    matrix_form : bool, default False
        Whether to return dmi as a matrix instead of a vector form.

    Returns
    -------
    dmi : (3,) or (3, 3) :numpy:`ndarray`
        Antisymmetic exchange (DMI). If ``matrix_form == True``, then return a matrix.

    Notes
    -----

    Antisymmetric anisotropy is defined as

    .. math::

        \boldsymbol{J}_{A}
        =
        \dfrac{\boldsymbol{J}
        -
        \boldsymbol{J}.T}{2}
        =
        \begin{pmatrix}
            0 & D^z & -D^y \\
            -D^z & 0 & D^x \\
            D^y & -D^x & 0 \\
        \end{pmatrix}


    In a vector form it is

    .. math::

        \boldsymbol{D}
        =
        \begin{pmatrix}
        D^x \\
        D^y \\
        D^z \\
        \end{pmatrix}

    Examples
    --------

    .. doctest::

        >>> import magnopy
        >>> matrix = [[1, 3, 0], [-1, -2, 3], [0, 3, 9]]
        >>> magnopy._spinham.get_dmi(matrix)
        array([0., 0., 2.])
        >>> magnopy._spinham.get_dmi(matrix, matrix_form = True)
        array([[ 0.,  2.,  0.],
               [-2.,  0.,  0.],
               [ 0.,  0.,  0.]])
    """

    matrix = np.array(matrix)

    asymm_matrix = (matrix - matrix.T) / 2

    if matrix_form:
        return asymm_matrix

    return np.array(
        [asymm_matrix[1][2], asymm_matrix[2][0], asymm_matrix[0][1]],
        dtype=float,
    )


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import sqrt

import numpy as np

from magnopy._diagonalization import solve_via_colpa
from magnopy._exceptions import ColpaFailed
from magnopy._local_rf import span_local_rfs
from magnopy._spinham._convention import Convention

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


class LSWT:
    r"""
    Linear Spin Wave theory.

    It is created from some spin Hamitlonian and set of direction vectors,
    that define the orientation of spins.

    Parameters
    ----------
    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian.
    spin_directions : (M, 3) |array-like|_
        Directions of spin vectors. Only directions of vectors are used, modulus is
        ignored.
        If spin Hamiltonian contains non-magnetic atom, then only the spin directions
        for the magnetic atoms are expected. The order of spin directions is the same as
        the order of magnetic atoms in ``spinham.atoms.spins``.

    Attributes
    ----------
    z : (M, 3) :numpy:`ndarray`
        Spin directions.
    p : (M, 3) :numpy:`ndarray`
        Hybridized x and y components of the local coordinate system.
    M : int
        Number of spins in the unit cell
    cell : (3, 3) :numpy:`ndarray`
        Unit cell. Rows are vectors, columns are cartesian components.
    spins : (M, ) :numpy:`ndarray`
        Spin values.

    Notes
    -----
    If spin Hamiltonian contains three atoms Cr1, Br, Cr3 in that order. Assume that two
    atoms are magnetic (Cr1 and Cr2), one atom is not (Br). Then ``spin_directions`` is
    a (2, 3) array with ``spin_directions[0]`` being the direction for spin of Cr1 and
    ``spin_directions[1]`` being the direction of spin for Cr2.
    """

    def __init__(self, spinham, spin_directions):
        spin_directions = np.array(spin_directions, dtype=float)
        spin_directions /= np.linalg.norm(spin_directions, axis=1)[:, np.newaxis]

        x, y, self.z = span_local_rfs(
            directional_vectors=spin_directions, hybridize=False
        )
        self.p = x + 1j * y

        self.spins = np.array(spinham.magnetic_atoms.spins, dtype=float)

        initial_convention = spinham.convention

        spinham.convention = initial_convention.get_modified(
            spin_normalized=False, multiple_counting=True
        )

        self.M = spinham.M
        self.cell = spinham.cell

        ########################################################################
        #                    Renormalized one-spin parameter                   #
        ########################################################################
        self._J1 = np.zeros((self.M, 3), dtype=float)

        # One spin
        for alpha, parameter in spinham.p1:
            alpha = spinham.index_map[alpha]
            self._J1[alpha] = self._J1[alpha] + spinham.convention.c1 * parameter

        # Two spins & one site
        for alpha, parameter in spinham.p21:
            alpha = spinham.index_map[alpha]
            self._J1[alpha] = self._J1[alpha] + (
                2
                * spinham.convention.c21
                * np.einsum("ij,j->i", parameter, self.z[alpha])
                * self.spins[alpha]
            )

        # Two spins & two sites
        for alpha, beta, _, parameter in spinham.p22:
            alpha = spinham.index_map[alpha]
            beta = spinham.index_map[beta]
            self._J1[alpha] = self._J1[alpha] + (
                2
                * spinham.convention.c22
                * (parameter @ self.z[beta])
                * self.spins[beta]
            )

        # Three spins & one site
        for alpha, parameter in spinham.p31:
            alpha = spinham.index_map[alpha]
            self._J1[alpha] = self._J1[alpha] + (
                3
                * spinham.convention.c31
                * np.einsum("iju,j,u->i", parameter, self.z[alpha], self.z[alpha])
                * self.spins[alpha] ** 2
            )

        # Three spins & two sites
        for alpha, beta, _, parameter in spinham.p32:
            alpha = spinham.index_map[alpha]
            beta = spinham.index_map[beta]
            self._J1[alpha] = self._J1[alpha] + (
                3
                * spinham.convention.c32
                * np.einsum("iju,j,u->i", parameter, self.z[alpha], self.z[beta])
                * self.spins[alpha]
                * self.spins[beta]
            )

        # Three spins & three sites
        for alpha, beta, gamma, _, _, parameter in spinham.p33:
            alpha = spinham.index_map[alpha]
            beta = spinham.index_map[beta]
            gamma = spinham.index_map[gamma]
            self._J1[alpha] = self._J1[alpha] + (
                3
                * spinham.convention.c33
                * np.einsum("iju,j,u->i", parameter, self.z[beta], self.z[gamma])
                * self.spins[beta]
                * self.spins[gamma]
            )

        # Four spins & one site
        for alpha, parameter in spinham.p41:
            alpha = spinham.index_map[alpha]
            self._J1[alpha] = self._J1[alpha] + (
                4
                * spinham.convention.c41
                * np.einsum(
                    "ijuv,j,u,v->i",
                    parameter,
                    self.z[alpha],
                    self.z[alpha],
                    self.z[alpha],
                )
                * self.spins[alpha] ** 3
            )

        # Four spins & two sites (1+3)
        for alpha, beta, _, parameter in spinham.p421:
            alpha = spinham.index_map[alpha]
            beta = spinham.index_map[beta]
            self._J1[alpha] = self._J1[alpha] + (
                4
                * spinham.convention.c421
                * np.einsum(
                    "ijuv,j,u,v->i",
                    parameter,
                    self.z[alpha],
                    self.z[alpha],
                    self.z[beta],
                )
                * self.spins[alpha] ** 2
                * self.spins[beta]
            )

        # Four spins & two sites (2+2)
        for alpha, beta, _, parameter in spinham.p422:
            alpha = spinham.index_map[alpha]
            beta = spinham.index_map[beta]
            self._J1[alpha] = self._J1[alpha] + (
                4
                * spinham.convention.c422
                * np.einsum(
                    "ijuv,j,u,v->i",
                    parameter,
                    self.z[alpha],
                    self.z[beta],
                    self.z[beta],
                )
                * self.spins[alpha]
                * self.spins[beta] ** 2
            )

        # Four spins & three sites
        for alpha, beta, gamma, _, _, parameter in spinham.p43:
            alpha = spinham.index_map[alpha]
            beta = spinham.index_map[beta]
            gamma = spinham.index_map[gamma]
            self._J1[alpha] = self._J1[alpha] + (
                4
                * spinham.convention.c43
                * np.einsum(
                    "ijuv,j,u,v->i",
                    parameter,
                    self.z[alpha],
                    self.z[beta],
                    self.z[gamma],
                )
                * self.spins[alpha]
                * self.spins[beta]
                * self.spins[gamma]
            )

        # Four spins & four sites
        for alpha, beta, gamma, epsilon, _, _, _, parameter in spinham.p44:
            alpha = spinham.index_map[alpha]
            beta = spinham.index_map[beta]
            gamma = spinham.index_map[gamma]
            epsilon = spinham.index_map[epsilon]
            self._J1[alpha] = self._J1[alpha] + (
                4
                * spinham.convention.c44
                * np.einsum(
                    "ijuv,j,u,v->i",
                    parameter,
                    self.z[beta],
                    self.z[gamma],
                    self.z[epsilon],
                )
                * self.spins[beta]
                * self.spins[gamma]
                * self.spins[epsilon]
            )

        ########################################################################
        #                   Renormalized two-spins parameter                   #
        ########################################################################
        self._J2 = {}

        # First - terms with delta in from of them

        # Two spins & one site
        for alpha, parameter in spinham.p21:
            alpha = spinham.index_map[alpha]
            if (0, 0, 0) not in self._J2:
                self._J2[(0, 0, 0)] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[(0, 0, 0)][alpha, alpha] += 2 * spinham.convention.c21 * parameter

        # Three spins & one site
        for alpha, parameter in spinham.p31:
            alpha = spinham.index_map[alpha]
            if (0, 0, 0) not in self._J2:
                self._J2[(0, 0, 0)] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[(0, 0, 0)][alpha, alpha] += (
                3
                * spinham.convention.c31
                * np.einsum("iju,u->ij", parameter, self.z[alpha])
                * self.spins[alpha]
            )

        # Four spins & one site
        for alpha, parameter in spinham.p41:
            alpha = spinham.index_map[alpha]
            if (0, 0, 0) not in self._J2:
                self._J2[(0, 0, 0)] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[(0, 0, 0)][alpha, alpha] += (
                6
                * spinham.convention.c41
                * np.einsum("ijuv,u,v->ij", parameter, self.z[alpha], self.z[alpha])
                * self.spins[alpha] ** 2
            )

        # Then all other parameters

        # Two spins & two sites
        for alpha, beta, nu, parameter in spinham.p22:
            alpha = spinham.index_map[alpha]
            beta = spinham.index_map[beta]
            if nu not in self._J2:
                self._J2[nu] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[nu][alpha, beta] += spinham.convention.c22 * parameter

        # Three spins & two sites
        for alpha, beta, nu, parameter in spinham.p32:
            alpha = spinham.index_map[alpha]
            beta = spinham.index_map[beta]
            if nu not in self._J2:
                self._J2[nu] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[nu][alpha, beta] += (
                3
                * spinham.convention.c32
                * np.einsum("iuj,u->ij", parameter, self.z[alpha])
                * self.spins[alpha]
            )

        # Three spins & three sites
        for alpha, beta, gamma, nu, _, parameter in spinham.p33:
            alpha = spinham.index_map[alpha]
            beta = spinham.index_map[beta]
            gamma = spinham.index_map[gamma]
            if nu not in self._J2:
                self._J2[nu] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[nu][alpha, beta] += (
                3
                * spinham.convention.c33
                * np.einsum("iju,u->ij", parameter, self.z[gamma])
                * self.spins[gamma]
            )

        # Four spins & two sites (1+3)
        for alpha, beta, nu, parameter in spinham.p421:
            alpha = spinham.index_map[alpha]
            beta = spinham.index_map[beta]
            if nu not in self._J2:
                self._J2[nu] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[nu][alpha, beta] += (
                6
                * spinham.convention.c421
                * np.einsum("iuvj,u,v->ij", parameter, self.z[alpha], self.z[alpha])
                * self.spins[alpha] ** 2
            )

        # Four spins & two sites (2+2)
        for alpha, beta, nu, parameter in spinham.p422:
            alpha = spinham.index_map[alpha]
            beta = spinham.index_map[beta]
            if nu not in self._J2:
                self._J2[nu] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[nu][alpha, beta] += (
                6
                * spinham.convention.c422
                * np.einsum("iujv,u,v->ij", parameter, self.z[alpha], self.z[beta])
                * self.spins[alpha]
                * self.spins[beta]
            )

        # Four spins & three sites
        for alpha, beta, gamma, nu, _, parameter in spinham.p43:
            alpha = spinham.index_map[alpha]
            beta = spinham.index_map[beta]
            gamma = spinham.index_map[gamma]
            if nu not in self._J2:
                self._J2[nu] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[nu][alpha, beta] += (
                6
                * spinham.convention.c43
                * np.einsum("iujv,u->ij", parameter, self.z[alpha], self.z[gamma])
                * self.spins[alpha]
                * self.spins[gamma]
            )

        # Four spins & four sites
        for alpha, beta, gamma, epsilon, nu, _, _, parameter in spinham.p44:
            alpha = spinham.index_map[alpha]
            beta = spinham.index_map[beta]
            gamma = spinham.index_map[gamma]
            epsilon = spinham.index_map[epsilon]
            if nu not in self._J2:
                self._J2[nu] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[nu][alpha, beta] += (
                6
                * spinham.convention.c44
                * np.einsum("ijuv,u->ij", parameter, self.z[gamma], self.z[epsilon])
                * self.spins[gamma]
                * self.spins[epsilon]
            )

        spinham.convention = initial_convention

        self.A1 = 0.5 * np.sum(self._J1 * self.z, axis=1)

        self.A2 = {}
        self.B2 = {}

        for nu in self._J2:
            self.A2[nu] = 0.5 * np.einsum(
                "abij,a,b,ai,bj->ab",
                self._J2[nu],
                np.sqrt(self.spins),
                np.sqrt(self.spins),
                self.p,
                np.conjugate(self.p),
            )
            self.B2[nu] = 0.5 * np.einsum(
                "abij,a,b,ai,bj->ab",
                self._J2[nu],
                np.sqrt(self.spins),
                np.sqrt(self.spins),
                np.conjugate(self.p),
                np.conjugate(self.p),
            )

    @property
    def E_2(self) -> float:
        r"""
        Correction to the ground state energy that arises from the LSWT.

        Returns
        -------
        E_2 : float
        """

        return float(0.5 * np.sum(self._J1 * self.z))

    @property
    def O(self):
        r"""
        Coefficient before the one-operator terms.

        Returns
        -------
        O : (M, ) :numpy:`ndarray`
            ``complex``.

        Notes
        -----
        Before the diagonalization, the magnon Hamiltonian has the form

        .. math::

            \mathcal{H}
            =
            \dots
            +
            \sqrt{N}
            \sum_{\alpha}
            \Bigl(
            O_{\alpha}
            a_{\alpha}(\boldsymbol{0})
            +
            \overline{O_{\alpha}}
            a^{\dagger}_{\alpha}(\boldsymbol{0})
            \Bigr)
            +
            \dots

        where overline denotes complex conjugation. This function computes the
        coefficients :math:`O_{\alpha}`.
        """

        return np.einsum(
            "a,ai,ai->a",
            np.sqrt(self.spins) / np.sqrt(2),
            np.conjugate(self.p),
            self._J1,
        )

    def A(self, k, relative=False):
        r"""
        Part of the Grand dynamical matrix.

        Parameters
        ----------
        k : (3,) |array-like|_
            Reciprocal vector
        relative : bool, default False
            If ``relative=True``, then ``k`` is interpreted as given relative to the
            reciprocal unit cell. Otherwise it is interpreted as given in absolute
            coordinates.

        Returns
        -------
        A : (M, M) :numpy:`ndarray`
            :math:`A_{\alpha\beta}(\boldsymbol{k})`.

        Notes
        -----
        Before the diagonalization, the magnon Hamiltonian has the form

        .. math::

            \mathcal{H}
            =
            \dots
            +
            \sum_{\boldsymbol{k}, \alpha}
            \boldsymbol{\mathcal{A}}(\boldsymbol{k})^{\dagger}
            \begin{pmatrix}
            \boldsymbol{A}(\boldsymbol{k}) & \boldsymbol{B}^{\dagger}(\boldsymbol{k}) \\
            \boldsymbol{B}(\boldsymbol{k}) & \overline{\boldsymbol{A}(-\boldsymbol{k})}
            \end{pmatrix}
            \boldsymbol{\mathcal{A}}(\boldsymbol{k})

        where

        .. math::
            \boldsymbol{\mathcal{A}}(\boldsymbol{k})
            =
            \begin{pmatrix}
            a_1(\boldsymbol{k}),
            \dots,
            a_M(\boldsymbol{k}),
            a^{\dagger}_1(-\boldsymbol{k}),
            \dots,
            a^{\dagger}_M(-\boldsymbol{k}),
            \end{pmatrix}

        This function computes the matrix :math:`\boldsymbol{A}(\boldsymbol{k})`.
        """

        k = np.array(k)

        result = np.zeros((self.M, self.M), dtype=complex)

        for nu in self.A2:
            if relative:
                phase = 2 * np.pi * (k @ nu)
            else:
                phase = k @ (nu @ self.cell)
            result = result + self.A2[nu] * np.exp(1j * phase)

        result = result - np.diag(self.A1)

        return result

    def B(self, k, relative=False):
        r"""
        Part of the Grand dynamical matrix.

        Parameters
        ----------
        k : (3,) |array-like|_
            Reciprocal vector
        relative : bool, default False
            If ``relative=True``, then ``k`` is interpreted as given relative to the
            reciprocal unit cell. Otherwise it is interpreted as given in absolute
            coordinates.

        Returns
        -------
        B : (M, M) :numpy:`ndarray`
            :math:`B_{\alpha\beta}(\boldsymbol{k})`.

        Notes
        -----
        Before the diagonalization, the magnon Hamiltonian has the form

        .. math::

            \mathcal{H}
            =
            \dots
            +
            \sum_{\boldsymbol{k}, \alpha}
            \boldsymbol{\mathcal{A}}(\boldsymbol{k})^{\dagger}
            \begin{pmatrix}
            \boldsymbol{A}(\boldsymbol{k}) & \boldsymbol{B}^{\dagger}(\boldsymbol{k}) \\
            \boldsymbol{B}(\boldsymbol{k}) & \overline{\boldsymbol{A}(-\boldsymbol{k})}
            \end{pmatrix}
            \boldsymbol{\mathcal{A}}(\boldsymbol{k})

        where

        .. math::
            \boldsymbol{\mathcal{A}}(\boldsymbol{k})
            =
            \begin{pmatrix}
            a_1(\boldsymbol{k}),
            \dots,
            a_M(\boldsymbol{k}),
            a^{\dagger}_1(-\boldsymbol{k}),
            \dots,
            a^{\dagger}_M(-\boldsymbol{k}),
            \end{pmatrix}

        This function computes the matrix :math:`\boldsymbol{B}(\boldsymbol{k})`.
        """

        k = np.array(k)

        result = np.zeros((self.M, self.M), dtype=complex)

        for nu in self.B2:
            if relative:
                phase = 2 * np.pi * (k @ nu)
            else:
                phase = k @ (nu @ self.cell)
            result = result + self.B2[nu] * np.exp(1j * phase)

        result = result

        return result

    def GDM(self, k, relative=False):
        r"""
        Grand dynamical matrix.

        Parameters
        ----------
        k : (3,) |array-like|_
            Reciprocal vector
        relative : bool, default False
            If ``relative=True``, then ``k`` is interpreted as given relative to the
            reciprocal unit cell. Otherwise it is interpreted as given in absolute
            coordinates.

        Returns
        -------
        gdm : (2M, 2M) :numpy:`ndarray`
            Gran dynamical matrix.

        Notes
        -----
        Before the diagonalization, the magnon Hamiltonian has the form

        .. math::

            \mathcal{H}
            =
            \dots
            +
            \sum_{\boldsymbol{k}, \alpha}
            \boldsymbol{\mathcal{A}}(\boldsymbol{k})^{\dagger}
            \begin{pmatrix}
            \boldsymbol{A}(\boldsymbol{k}) & \boldsymbol{B}^{\dagger}(\boldsymbol{k}) \\
            \boldsymbol{B}(\boldsymbol{k}) & \overline{\boldsymbol{A}(-\boldsymbol{k})}
            \end{pmatrix}
            \boldsymbol{\mathcal{A}}(\boldsymbol{k})

        where

        .. math::
            \boldsymbol{\mathcal{A}}(\boldsymbol{k})
            =
            \begin{pmatrix}
            a_1(\boldsymbol{k}),
            \dots,
            a_M(\boldsymbol{k}),
            a^{\dagger}_1(-\boldsymbol{k}),
            \dots,
            a^{\dagger}_M(-\boldsymbol{k}),
            \end{pmatrix}

        This function computes the grand dynamical matrix
        :math:`\boldsymbol{D}(\boldsymbol{k})`

        .. math::

            \boldsymbol{D}(\boldsymbol{k})
            =
            \begin{pmatrix}
            \boldsymbol{A}(\boldsymbol{k}) & \boldsymbol{B}^{\dagger}(\boldsymbol{k}) \\
            \boldsymbol{B}(\boldsymbol{k}) & \overline{\boldsymbol{A}(-\boldsymbol{k})}
            \end{pmatrix}
        """

        k = np.array(k, dtype=float)

        A = self.A(k=k, relative=relative)
        A_m = self.A(k=-k, relative=relative)

        B = self.B(k=k, relative=relative)

        left = np.concatenate((A, B), axis=0)
        right = np.concatenate((np.conjugate(B).T, np.conjugate(A_m)), axis=0)
        gdm = np.concatenate((left, right), axis=1)

        return gdm

    def diagonalize(self, k, relative=False):
        r"""
        Diagonalize the Hamiltonian for the given ``k`` point and return all possible
        quantity at once.

        Parameters
        ----------
        k : (3,) |array-like|_
            Reciprocal vector
        relative : bool, default False
            If ``relative=True``, then ``k`` is interpreted as given relative to the
            reciprocal unit cell. Otherwise it is interpreted as given in absolute
            coordinates.

        Returns
        -------
        omegas : (M, ) :numpy:`ndarray`
            Array of omegas. Note, that data type is complex. If the ground state is correct,
            then the complex part should be zero.
        delta : float
            Constant energy term that results from diagonalization. Note, that data type is complex. If the ground state is correct,
            then the complex part should be zero.
        G_minus_one : (M, 2M) :numpy:`ndarray`
            Transformation matrix from the original boson operators.
            Note that this function returns :math:`(\mathcal{G})^{-1}` for convenience.

            .. math::

                \begin{pmatrix}
                    b_1(\boldsymbol{k}),
                    \dots,
                    b_M(\boldsymbol{k}),
                \end{pmatrix}
                =
                (\mathcal{G})^{-1}

                \begin{pmatrix}
                    a_1(\boldsymbol{k}),
                    \dots,
                    a_M(\boldsymbol{k}),
                    a^{\dagger}_1(-\boldsymbol{k}),
                    \dots,
                    a^{\dagger}_M(-\boldsymbol{k}),
                \end{pmatrix}
        """

        k_plus = np.array(k)
        k_minus = -k_plus

        GDM_plus = self.GDM(k_plus, relative=relative)
        GDM_minus = self.GDM(k_minus, relative=relative)

        try:
            E_plus, G_plus = solve_via_colpa(GDM_plus, return_inverse=True)
            E_minus, _G_minus = solve_via_colpa(GDM_minus, return_inverse=True)
        except ColpaFailed:
            try:
                E_plus, G_plus = solve_via_colpa(-GDM_plus, return_inverse=True)
                E_minus, G_minus = solve_via_colpa(-GDM_minus, return_inverse=True)
            except ColpaFailed:
                try:
                    E_plus, G_plus = solve_via_colpa(
                        GDM_plus + (1e-8) * np.ones(GDM_plus.shape, dtype=float),
                        return_inverse=True,
                    )
                    E_minus, G_minus = solve_via_colpa(
                        GDM_minus + (1e-8) * np.ones(GDM_minus.shape, dtype=float),
                        return_inverse=True,
                    )
                except ColpaFailed:
                    return (
                        [np.nan for _ in range(self.M)],
                        np.nan,
                        [[np.nan for _ in range(2 * self.M)] for _ in range(self.M)],
                    )

        return (
            E_plus[: self.M].real + E_minus[self.M :],
            0.5 * (np.sum(E_plus[self.M :]) - np.sum(E_plus[: self.M])),
            G_plus[: self.M],
        )

    def omega(self, k, relative=False):
        r"""
        Parameters
        ----------
        k : (3,) |array-like|_
            Reciprocal vector
        relative : bool, default False
            If ``relative=True``, then ``k`` is interpreted as given relative to the
            reciprocal unit cell. Otherwise it is interpreted as given in absolute
            coordinates.

        Returns
        -------
        omegas : (M, ) :numpy:`ndarray`
            Array of omegas. Note, that data type is complex. If the ground state is correct,
            then the complex part should be zero.
        """

        return self.diagonalize(k=k, relative=relative)[0]

    def delta(self, k, relative=False):
        r"""
        Constant energy term of the diagonalized Hamiltonian.

        .. math::

            \sum_{\boldsymbol{k}}\Delta(\boldsymbol{k})

        Parameters
        ----------
        k : (3,) |array-like|_
            Reciprocal vector
        relative : bool, default False
            If ``relative=True``, then ``k`` is interpreted as given relative to the
            reciprocal unit cell. Otherwise it is interpreted as given in absolute
            coordinates.

        Returns
        -------
        delta : float
            Constant energy term that results from diagonalization. Note, that data type is complex. If the ground state is correct,
            then the complex part should be zero.
        """
        return self.diagonalize(k=k, relative=relative)[1]

    def G_minus_one(self, k, relative=False):
        r"""
        Inverse of the transformation matrix to the new bosonic operators.

        .. math::

            b_{\alpha}(\boldsymbol{k})
            =
            \sum_{\beta}
            (\mathcal{G}^{-1})_{\alpha, \beta}(\boldsymbol{k})
            \mathcal{A}_{\beta}(\boldsymbol{k})

        Parameters
        ----------
        k : (3,) |array-like|_
            Reciprocal vector
        relative : bool, default False
            If ``relative=True``, then ``k`` is interpreted as given relative to the
            reciprocal unit cell. Otherwise it is interpreted as given in absolute
            coordinates.

        Returns
        -------
        G_minus_one : (M, 2M) :numpy:`ndarray`
            Transformation matrix from the original boson operators.
            Note that this function returns :math:`(\mathcal{G})^{-1}` for convenience.
        """
        return self.diagonalize(k=k, relative=relative)[2]


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


R"""
Constants and custom units in the International system of units.
"""

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")

################################################################################
#                                    Units                                     #
################################################################################
# Length, in Metre
ANGSTROM = 1e-10
# Energy, in Joule
ELECTRON_VOLT = 1.602176634e-19
# Time, in Second
SECOND = 1
# Temperature, in Kelvin
KELVIN = 1
# Magnetic field, in Tesla
TESLA = 1


################################################################################
#                                  Constants                                   #
################################################################################
K_BOLTZMANN = 1.380649e-23  # Joule / Kelvin
BOHR_RADIUS = 5.29177210903e-11  # Metre
BOHR_MAGNETON = 9.2740100783e-24  # Joule / Tesla
PLANK_CONSTANT = 6.62607015e-34  # Joule * Second
SPEED_OF_LIGHT = 299792458  # Metre / Second
RYDBERG_CONSTANT = 10973731.568160  # 1 / Metre
RYDBERG_ENERGY = PLANK_CONSTANT * SPEED_OF_LIGHT * RYDBERG_CONSTANT  # Joule
ELEMENTARY_CHARGE = 1.602176634e-19  # Coulomb

# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import logging

import numpy as np

from magnopy.energy.line_search import _line_search

_logger = logging.getLogger(__name__)

__all__ = []


def _bfgs(
    self,
    true_variables_0,
    hessian_0,
    tolerance,
    func,
    grad,
    update,
    to_x,
    update_difference,
):
    r"""
    Generalized optimization function. It's implementation is a little complicated as
    we need to track both general vector parameter :math:`\boldsymbol{x}` and the true
    variables of the energy function (i.e. spin directions, cone axis, spiral vector).

    Parameters
    ----------
    true_variables_0 : ...
        Initial guess for the true variables. Form and type depends on the
        ground state sub-type. The calls of ``to_x(true_variables_0)``,
        ``func(true_variables_0)``, ``grad(true_variables_0)`` have to be valid.
    hessian_0 : ...
        Initial guess for the hessian matrix.
    tolerance : (N, ) :numpy:`ndarray`
        One dimensional array of tolerance targets for the optimization.
        Exact amount of elements depends on the ground state sub-type.
    func : callable
        Energy function. The call ``func(true_variables_0)`` has to be valid.
    grad : callable
        Gradient of the energy function. The call ``grad(true_variables_0)`` has
        to be valid.
    update : callable
        Function that update the true variables to the new state defined by the
        :math:`\boldsymbol{x}` vector. The call ``update(true_variables_0, x)``
        has to be valid.
    to_x : callable
        Function that converts true variables to the generalized vector argument. The
        call ``to_x(true_variables_0)`` has to be valid.
    update_difference : callable
        Function that computes the difference between two consecutive steps of the
        optimization. The call ``update_difference(func_k, func_kp1, grad_k, grad_kp1)``
        has to be valid. Where ``func_k = func(true_variables_k)`` and
        ``grad_k = grad(true_variables_k)``.

    Returns
    -------
    true_variables : ...
        Optimized true variables.
    """

    difference = tolerance.copy() * 2
    true_variables_k = true_variables_0.copy()
    x_k = to_x(true_variables_k)
    hessian_size = x_k.size
    # hessian = np.eye(hessian_size, dtype=float)

    hessian = np.linalg.inv(hessian_0)

    func_k = func(true_variables_k)
    grad_k = grad(true_variables_k)

    step_counter = 0
    first_run = True
    while (difference > tolerance).any():
        search_direction = -hessian @ grad_k

        alphas = np.linspace(0, 1.5, 300)

        energies = []
        for alpha in alphas:
            tmp_x = x_k + alpha * search_direction
            tmp_tv = update(true_variables_k, tmp_x)

            energies.append(func(tmp_tv))

        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(figsize=(20, 4))

        ax.plot(alphas, energies, lw=2, color="black")

        # Fix it tomorrow
        alpha_k = _line_search(
            x_k=x_k,
            true_variables_k=true_variables_k,
            search_direction=search_direction,
            func=func,
            grad=grad,
            update=update,
            func_0=func_k,
            grad_0=grad_k,
        )

        ax.vlines(
            [0, alpha_k], 0, 1, lw=0.5, color="grey", transform=ax.get_xaxis_transform()
        )
        ax.hlines(
            energies[0], 0, 1, lw=0.5, color="grey", transform=ax.get_yaxis_transform()
        )
        ax.set_title(
            R"$\alpha = " f"{alpha_k:.8f}$",
            fontsize=20,
        )

        plt.show()
        plt.close()

        x_kp1 = x_k + alpha_k * search_direction
        true_variables_kp1 = update(true_variables_k, x_kp1)

        grad_kp1 = grad(true_variables_kp1)
        func_kp1 = func(true_variables_kp1)

        difference = update_difference(func_k, func_kp1, grad_kp1)

        print(difference)
        step_counter += 1

        if (difference < tolerance).all():
            break

        s_k = x_kp1 - x_k
        y_k = grad_kp1 - grad_k

        rho_k = 1 / (y_k @ s_k)

        EYE = np.eye(hessian_size, dtype=float)
        OUTER = np.outer(y_k, s_k)

        # if first_run:
        #     first_run = False
        #     hessian = (s_k @ y_k) / (y_k @ y_k) * hessian

        hessian = (EYE - rho_k * OUTER.T) @ hessian @ (
            EYE - rho_k * OUTER
        ) + rho_k * np.outer(s_k, s_k)

        grad_k = grad_kp1
        func_k = func_kp1
        x_k = to_x(true_variables_kp1)
        true_variables_k = true_variables_kp1

    _logger.info(f"Optimization finished in {step_counter} steps.")

    return true_variables_kp1
